!#####################################################################
!!
!!  File  input_rcl.f90
!!  is part of RECOLA (REcursive Computation of One Loop Amplitudes)
!!
!!  Copyright (C) 2015-2019   Stefano Actis, Ansgar Denner,
!!                            Lars Hofer, Jean-Nicolas Lang,
!!                            Andreas Scharf, Sandro Uccirati
!!
!!  RECOLA is licenced under the GNU GPL version 3,
!!         see COPYING for details.
!!
!#####################################################################

  module input_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  use globals_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  implicit none

!---------------------------------------------------------------------
! Physical variables
!---------------------------------------------------------------------

! Pole Masses:
! The Z,W masses and width are computed from the on-shell masses
! and width:
! mass_z_onshell =  91.1876d0, width_z_onshell = 2.4952d0,
! mass_w_onshell =  80.385d0,  width_w_onshell = 2.085d0,
  real(dp) :: mass_z  =  91.153480619182744d0, width_z  = 2.4942663787728243d0
  real(dp) :: mass_w  =  80.357973609877547d0, width_w  = 2.0842989982782196d0
  real(dp) :: mass_h  = 125.d0,                width_h  = 0.d0
  real(dp) :: mass_el =   0.d0
  real(dp) :: mass_mu =   0.d0,                width_mu = 0.d0
  real(dp) :: mass_ta =   0.d0,                width_ta = 0.d0
  real(dp) :: mass_u  =   0.d0
  real(dp) :: mass_d  =   0.d0
  real(dp) :: mass_c  =   0.d0,                width_c  = 0.d0
  real(dp) :: mass_s  =   0.d0
  real(dp) :: mass_t  = 173.2d0,               width_t  = 0.d0
  real(dp) :: mass_b  =   0.d0,                width_b  = 0.d0

! Regularization of collinear singularities:
! Dimensional regularization is used for collinear singularities of
! massless fermions.
! Massive fermions are treated according to the variables "light_x"
! (x = el,mu,ta,u,d,c,s,t,b):
! - light_x = .true.
!   Fermion "x" is light (terms proportional to the mass are
!   discarded). Mass regularization is used for collinear divergences
!   with mass regulator given by the pole mass of fermion "x".
! - light_x = .false.
!   Fermion "x" is heavy (full mass dependence is kept).
! If the mass of fermion "x" is zero, "light_x" is irrelevant.
  logical :: light_el = .true.
  logical :: light_mu = .true.
  logical :: light_ta = .true.
  logical :: light_u  = .true.
  logical :: light_d  = .true.
  logical :: light_c  = .true.
  logical :: light_s  = .true.
  logical :: light_t  = .false.
  logical :: light_b  = .true.

! Regularization of soft singularities:
! "reg_soft" defines the regularization of soft singularities
! generated by massless photons and gluons:
! - reg_soft = 1:
!   dimensional regularization is used for soft singularities
!   ("lambda" irrelevant).
! - reg_soft = 2:
!   mass regularization is used for soft singularities with mass
!   regulator "lambda".
! REMARK: mass regularization for soft singularities (reg_soft = 2)
!         is allowed only in combination with mass regularization
!         for collinear singularities of light fermions. Therefore
!         if reg_soft = 2, no massless fermions are allowed
!         (and the code stops).
  integer :: reg_soft = 1;   real(dp) :: lambda = 100d0

! UV and IR dimensional regulators:
! "DeltaUV"  is the coefficient of the UV pole in the MSbar scheme
!            1/ep - EulerConstant + ln(4*pi)
! "muUV"     is the UV scale
! "DeltaIR"  is the coefficient of the IR single pole
! "DeltaIR2" is the coefficient of the IR double pole
! "muIR"     is the IR scale
  real(dp) :: DeltaUV = 0d0,                 muUV = 100d0
  real(dp) :: DeltaIR = 0d0, DeltaIR2 = 0d0, muIR = 100d0

! alpha_s, renormalization scale and flavour scheme:
! "Qren" and "als" are the initialization values for the
! renormalization scale and for alpha_s.
! "Nfren" is the number of active flavours for alpha_s renormalization.
! "als" should correspond to the physical value of alpha_s at the
! scale "Qren" computed in the "Nfren"-flavour scheme.
! Accepted values for "Nfren" are -1,3,4,5,6
! Nfren = -1      -> all quarks with mass less than "Qren" are active
!                    flavours
! Nfren = 3,4,5,6 -> the 3,4,5,6 lightest quarks are the active
!                    flavours
  real(dp) :: als   = 0.118d0
  real(dp) :: Qren  = 91.1876d0
  integer  :: Nfren = 5

! alpha, Gf and scheme for renormalization of alpha:
! - ew_reno_scheme = 1 -> gfermi scheme:
!   alpha is determined from G_F ("gf").
!   Universal corrections enhanced by large mass ratios are resummed.
!   ("al0", "alZ" are irrelevant)
! - ew_reno_scheme = 2 -> alpha0 scheme:
!   alpha is renormalized at p^2 = 0 ("al0").
!   No resummation of universal corrections
!   ("gf", "alZ" are irrelevant)
! - ew_reno_scheme = 3 -> alphaZ scheme:
!   alpha is renormalized at p^2 = mz^2 ("alZ").
!   No resummation of universal corrections
!   ("gf", "al0" are irrelevant)
  integer  :: ew_reno_scheme = 1
  real(dp) :: gf  = 1.16637d-5
  real(dp) :: al0 = 1d0/137.035999679d0
  real(dp) :: alZ = 1d0/128.936d0

  ! Treatment of W and Z masses in the computation of alpha_gf in
  ! the gfermi scheme:
  ! masses_in_gf_reno_scheme = 0:
  !  alpha_GF = G_F*sqrt(2)/pi*Re(MW^2)*( 1 - Re(MW^2)/Re(MZ^2) )
  ! masses_in_gf_reno_scheme = 1:
  !  alpha_GF = G_F*sqrt(2)/pi*abs( MW^2*( 1 - MW^2/MZ^2) )
  integer  :: masses_in_gf_reno_scheme = 1

! Complex mass scheme:
! - complex_mass_scheme = 1:
!   The complex mass scheme of hep-ph/0505042 is used.
! - complex_mass_scheme = 0:
!   The on-shell scheme is used, i.e. just the real part of the
!   self-energies are taken for the computation of the counterterms.
!   The imaginary part of the masses are kept just in the
!   propagators (also in fermion propagator i_/(pslash-m), m is kept
!   complex).
  integer :: complex_mass_scheme = 1

! Self-energies of resonant particles
  logical :: resSE = .true.

! Dynamic settings
  integer :: dynamic_settings = 0

! Switch for the correction of external momenta to restore the
! mass-shell conditions and momentum conservation
  logical :: momenta_correction = .true.

!---------------------------------------------------------------------
! Latex
!---------------------------------------------------------------------

! Generate a .tex file to draw the branches of the processes:
! 0 -> the .tex file is not generated.
! 1 -> the .tex file is generated without colour structures.
! 2 -> the .tex file is generated with colour structures.
  integer :: draw = 0

!---------------------------------------------------------------------
! Print options
!---------------------------------------------------------------------

! Print amplitudes
! writeMat = 0 -> nothing is printed.
! writeMat = 1 -> The non vanishining amplitudes A(c,h) are printed
!                 at LO and NLO (if NLO is computed) for each coulour
!                 structure "c" and helicity "h"
! writeMat = 2 -> In addition to writeMat=1, the NLO (if computed)
!                 contributions to A(c,h) from bare-loop, counterterms
!                 and rational part are also printed.
  integer :: writeMat  = 0

! Print summed squared amplitude
! writeMat2 = 0 -> nothing is printed.
! writeMat2 = 1 -> The non vanishining squared amplitudes A2
!                  summed/averaged over colour and helicities are
!                  printed at LO and NLO (if NLO is computed).
! writeMat2 = 2 -> In addition to writeMat2=1, the NLO (if computed)
!                  contributions to A2 from bare-loop, counterterms
!                  and rational part are also printed.
! writeMat2 = 3 -> In addition to writeMat2=1, also the squared
!                  amplitudes A2h summed/averaged over colour are
!                  printed at LO and NLO (if NLO is computed) for
!                  each helicity "h".
  integer :: writeMat2 = 0

! Print matrix elements squared with colour correlation (if computed)
! writeCor = 0 -> nothing is printed.
! writeCor = 1 -> The non vanishining colour-correlated Born squared
!                 amplitudes A2c(n,m) are printed for each emitter
!                 "n" and spectator "m".
  integer :: writeCor = 0

! Print RAM needed in the present run of Recola:
! writeRAM = 0 -> nothing is printed.
! writeRAM = 1 -> RAM permanently used by process generation and process
!                 computation is printed.
! writeRAM = 2 -> In addition to writeRAM=1, RAM temporally used by
!                 process generation is printed.
  integer :: writeRAM = 0

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  contains

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_z_rcl (m,g)

  ! This subroutine sets the pole mass ("mass_z") and width
  ! ("width_z") of the Z boson to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.le.0d0.or.g.lt.0d0) then
    if (warnings(200).le.warning_limit) then
      warnings(200) = warnings(200) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 200: set_pole_mass_z_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(200)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(201).le.warning_limit) then
      warnings(201) = warnings(201) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 201: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_z_rcl ', &
                           'with arguments'
      write(nx,*) '          m = ',m
      write(nx,*) '          g = ',g
      write(nx,*) '         has no effects.'
      write(nx,*)
      call toomanywarnings(201)
    endif
  endif

   mass_z = m
  width_z = g

  end subroutine set_pole_mass_z_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_onshell_mass_z_rcl (m,g)

  ! This subroutine sets the on-shell mass and width of the Z boson
  ! to "m" and "g" respectively. This is done by setting the pole mass
  ! "mass_z" to "m/sqrt(1+g^2/m^2)" and the pole width "width_z" to
  ! "g/sqrt(1+g^2/m^2)".

  real(dp), intent(in) :: m,g

  if (m.le.0d0.or.g.lt.0d0) then
    if (warnings(202).le.warning_limit) then
      warnings(202) = warnings(202) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 202: set_onshell_mass_z_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(202)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(203).le.warning_limit) then
      warnings(203) = warnings(203) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 203: Processes already generated.'
      write(nx,*) '             The call of set_onshell_mass_z_rcl ', &
                           'with arguments'
      write(nx,*) '          m = ',m
      write(nx,*) '          g = ',g
      write(nx,*) '         has no effects.'
      write(nx,*)
      call toomanywarnings(203)
    endif
  endif

   mass_z = m / sqrt ( 1d0 + g**2/m**2 )
  width_z = g / sqrt ( 1d0 + g**2/m**2 )

  end subroutine set_onshell_mass_z_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_w_rcl (m,g)

  ! This subroutine sets the pole mass ("mass_w") and width
  ! ("width_w") of the W boson to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.le.0d0.or.g.lt.0d0) then
    if (warnings(204).le.warning_limit) then
      warnings(204) = warnings(204) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 204: set_pole_mass_w_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(204)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(205).le.warning_limit) then
      warnings(205) = warnings(205) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 205: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_w_rcl ', &
                           'with arguments'
      write(nx,*) '              m = ',m
      write(nx,*) '              g = ',g
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(205)
    endif
  endif

   mass_w = m
  width_w = g

  end subroutine set_pole_mass_w_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_onshell_mass_w_rcl (m,g)

  ! This subroutine sets the on-shell mass and width of the W boson
  ! to "m" and "g" respectively. It is done by setting the pole mass
  ! "mass_w" to "m/sqrt(1+g^2/m^2)" and the pole width "width_w" to
  ! "g/sqrt(1+g^2/m^2)".

  real(dp), intent(in) :: m,g

  if (m.le.0d0.or.g.lt.0d0) then
    if (warnings(206).le.warning_limit) then
      warnings(206) = warnings(206) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 206: set_onshell_mass_w_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(206)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(207).le.warning_limit) then
      warnings(207) = warnings(207) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 207: Processes already generated.'
      write(nx,*) '             The call of set_onshell_mass_w_rcl ', &
                           'with arguments'
      write(nx,*) '              m = ',m
      write(nx,*) '              g = ',g
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(207)
    endif
  endif

   mass_w = m / sqrt ( 1d0 + g**2/m**2 )
  width_w = g / sqrt ( 1d0 + g**2/m**2 )

  end subroutine set_onshell_mass_w_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_h_rcl (m,g)

  ! This subroutine sets the pole mass ("mass_h") and width
  ! ("width_h") of the Higgs boson to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.le.0d0.or.g.lt.0d0) then
    if (warnings(208).le.warning_limit) then
      warnings(208) = warnings(208) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 208: set_pole_mass_h_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(208)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(209).le.warning_limit) then
      warnings(209) = warnings(209) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 209: Processes already generated.'
      write(nx,*) '            The call of set_pole_mass_h_rcl ', &
                           'with arguments'
      write(nx,*) '             m = ',m
      write(nx,*) '             g = ',g
      write(nx,*) '            has no effects.'
      write(nx,*)
      call toomanywarnings(209)
    endif
  endif

   mass_h = m
  width_h = g

  end subroutine set_pole_mass_h_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_electron_rcl (m)

  ! This subroutine sets pole mass "mass_el" of the electron to "m".

  real(dp), intent(in) :: m

  if (m.lt.0d0) then
    if (warnings(210).le.warning_limit) then
      warnings(210) = warnings(210) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 210: set_pole_mass_electron_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(210)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(211).le.warning_limit) then
      warnings(211) = warnings(211) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 211: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_electron_rcl ', &
                           'with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(211)
    endif
  endif

  mass_el = m

  end subroutine set_pole_mass_electron_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_muon_rcl (m,g)

  ! This subroutine sets pole mass ("mass_mu") and width ("width_mu")
  ! of the muon to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.lt.0d0.or.g.lt.0d0) then
    if (warnings(212).le.warning_limit) then
      warnings(212) = warnings(212) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 212: set_pole_mass_muon_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(212)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(213).le.warning_limit) then
      warnings(213) = warnings(213) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 213: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_muon_rcl ', &
                           'with arguments'
      write(nx,*) '              m = ',m
      write(nx,*) '              g = ',g
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(213)
    endif
  endif

   mass_mu = m
  width_mu = g

  end subroutine set_pole_mass_muon_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_tau_rcl (m,g)

  ! This subroutine sets pole mass ("mass_ta") and width ("width_ta")
  ! of the tau-lepton to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.lt.0d0.or.g.lt.0d0) then
    if (warnings(214).le.warning_limit) then
      warnings(214) = warnings(214) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 214: set_pole_mass_tau_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(214)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(215).le.warning_limit) then
      warnings(215) = warnings(215) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 215: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_tau_rcl ', &
                           'with arguments'
      write(nx,*) '              m = ',m
      write(nx,*) '              g = ',g
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(215)
    endif
  endif

   mass_ta = m
  width_ta = g

  end subroutine set_pole_mass_tau_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_up_rcl (m)

  ! This subroutine sets the pole mass "mass_u" of the up-quark to
  ! "m".

  real(dp), intent(in) :: m

  if (m.lt.0d0) then
    if (warnings(216).le.warning_limit) then
      warnings(216) = warnings(216) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 216: set_pole_mass_up_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(216)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(217).le.warning_limit) then
      warnings(217) = warnings(217) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 217: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_up_rcl ', &
                           'with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(217)
    endif
  endif

  mass_u = m

  end subroutine set_pole_mass_up_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_down_rcl (m)

  ! This subroutine sets the pole mass "mass_d" of the down-quark to
  ! "m".

  real(dp), intent(in) :: m

  if (m.lt.0d0) then
    if (warnings(218).le.warning_limit) then
      warnings(218) = warnings(218) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 218: set_pole_mass_down_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(218)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(219).le.warning_limit) then
      warnings(219) = warnings(219) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 219: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_down_rcl ', &
                           'with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(219)
    endif
  endif

  mass_d = m

  end subroutine set_pole_mass_down_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_charm_rcl (m,g)

  ! This subroutine sets the pole mass ("mass_c") and width
  ! ("width_c") of the charm-quark to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.lt.0d0.or.g.lt.0d0) then
    if (warnings(220).le.warning_limit) then
      warnings(220) = warnings(220) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 220: set_pole_mass_charm_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(220)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(221).le.warning_limit) then
      warnings(221) = warnings(221) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 221: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_charm_rcl ', &
                           'with arguments'
      write(nx,*) '              m = ',m
      write(nx,*) '              g = ',g
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(221)
    endif
  endif

   mass_c = m
  width_c = g

  end subroutine set_pole_mass_charm_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_strange_rcl (m)

  ! This subroutine sets the pole mass "mass_s" of the strange-quark
  ! to "m".

  real(dp), intent(in) :: m

  if (m.lt.0d0) then
    if (warnings(222).le.warning_limit) then
      warnings(222) = warnings(222) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 222: set_pole_mass_strange_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(222)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(223).le.warning_limit) then
      warnings(223) = warnings(223) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 223: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_strange_rcl ', &
                           'with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(223)
    endif
  endif

  mass_s = m

  end subroutine set_pole_mass_strange_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_top_rcl (m,g)

  ! This subroutine sets the pole mass ("mass_t") and width
  ! ("width_t") of the top-quark to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.lt.0d0.or.g.lt.0d0) then
    if (warnings(224).le.warning_limit) then
      warnings(224) = warnings(224) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 224: set_pole_mass_top_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(224)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(225).le.warning_limit) then
      warnings(225) = warnings(225) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 225: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_top_rcl ', &
                           'with arguments'
      write(nx,*) '              m = ',m
      write(nx,*) '              g = ',g
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(225)
    endif
  endif

   mass_t = m
  width_t = g

  end subroutine set_pole_mass_top_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_pole_mass_bottom_rcl (m,g)

  ! This subroutine sets the pole mass ("mass_b") and width
  ! ("width_b") of the bottom-quark to "m" and "g" respectively.

  real(dp), intent(in) :: m,g

  if (m.lt.0d0.or.g.lt.0d0) then
    if (warnings(226).le.warning_limit) then
      warnings(226) = warnings(226) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 226: set_pole_mass_bottom_rcl called with wrong arguments'
      write(nx,*)
      call toomanywarnings(226)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(227).le.warning_limit) then
      warnings(227) = warnings(227) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 227: Processes already generated.'
      write(nx,*) '             The call of set_pole_mass_bottom_rcl ', &
                           'with arguments'
      write(nx,*) '              m = ',m
      write(nx,*) '              g = ',g
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(227)
    endif
  endif

   mass_b = m
  width_b = g

  end subroutine set_pole_mass_bottom_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_fermions_rcl (m)

  ! This subroutine declares all massive fermions with mass smaller
  ! than or equal to "m" as light (i.e. their collinear singularities
  ! are regularized with their mass and terms proportional to the
  ! mass are discarded). Massive fermions with mass larger than "m"
  ! are heavy (i.e full mass dependence is kept). This subroutine
  ! has no effect on massless fermions (their collinear singularities
  ! are regularized dimensionally).

  real(dp), intent(in) :: m

  if (m.lt.0d0) then
    if (warnings(228).le.warning_limit) then
      warnings(228) = warnings(228) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 228: set_light_fermions_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(228)
    endif
    call istop (ifail,1)
  endif

  if (processes_generated) then
    if (warnings(229).le.warning_limit) then
      warnings(229) = warnings(229) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 229: Processes already generated.'
      write(nx,*) '             The call of set_light_fermions_rcl ', &
                           'with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(229)
    endif
  endif

  light_el = mass_el.le.m
  light_mu = mass_mu.le.m
  light_ta = mass_ta.le.m

  light_u = mass_u.le.m
  light_d = mass_d.le.m
  light_c = mass_c.le.m
  light_s = mass_s.le.m
  light_t = mass_t.le.m
  light_b = mass_b.le.m

  end subroutine set_light_fermions_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_electron_rcl

  ! This subroutine sets the electron as light. This means that, if
  ! the electron is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the electron is massless this subroutine has no
  ! effect (i.e. electron collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(230).le.warning_limit) then
      warnings(230) = warnings(230) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 230: Processes already generated.'
      write(nx,*) '             The call of set_light_electron_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(230)
    endif
  endif

  light_el = .true.

  end subroutine set_light_electron_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_muon_rcl

  ! This subroutine sets the muon as light. This means that, if the
  ! muon is massive, its collinear singularities are regularized with
  ! its mass and terms proportional to the mass are discarded. If the
  ! muon is massless this subroutine has no effect (i.e. muon
  ! collinear singularities are regularized dimensionally).

  if (processes_generated) then
    if (warnings(231).le.warning_limit) then
      warnings(231) = warnings(231) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 231: Processes already generated.'
      write(nx,*) '             The call of set_light_muon_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(231)
    endif
  endif

  light_mu = .true.

  end subroutine set_light_muon_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_tau_rcl

  ! This subroutine sets the tau-lepton as light. This means that, if
  ! the tau-lepton is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the tau-lepton is massless this subroutine has no
  ! effect (i.e. tau-lepton collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(232).le.warning_limit) then
      warnings(232) = warnings(232) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 232: Processes already generated.'
      write(nx,*) '             The call of set_light_tau_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(232)
    endif
  endif

  light_ta = .true.

  end subroutine set_light_tau_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_up_rcl

  ! This subroutine sets the up-quark as light. This means that, if
  ! the up-quark is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the up-quark is massless this subroutine has no
  ! effect (i.e. up-quark collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(233).le.warning_limit) then
      warnings(233) = warnings(233) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 233: Processes already generated.'
      write(nx,*) '             The call of set_light_up_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(233)
    endif
  endif

  light_u = .true.

  end subroutine set_light_up_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_down_rcl

  ! This subroutine sets the down quark as light. This means that, if
  ! the down-quark is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the down-quark is massless this subroutine has no
  ! effect (i.e. down-quark collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(234).le.warning_limit) then
      warnings(234) = warnings(234) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 234: Processes already generated.'
      write(nx,*) '             The call of set_light_down_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(234)
    endif
  endif

  light_d = .true.

  end subroutine set_light_down_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_charm_rcl

  ! This subroutine sets the charm quark as light. This means that,
  ! if the charm-quark is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the charm-quark is massless this subroutine has no
  ! effect (i.e. charm-quark collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(235).le.warning_limit) then
      warnings(235) = warnings(235) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 235: Processes already generated.'
      write(nx,*) '             The call of set_light_charm_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(235)
    endif
  endif

  light_c = .true.

  end subroutine set_light_charm_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_strange_rcl

  ! This subroutine sets the strange quark as light. This means that,
  ! if the strange-quark is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the strange-quark is massless this subroutine has
  ! no effect (i.e. strange-quark collinear singularities are
  ! regularized dimensionally).

  if (processes_generated) then
    if (warnings(236).le.warning_limit) then
      warnings(236) = warnings(236) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 236: Processes already generated.'
      write(nx,*) '             The call of set_light_strange_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(236)
    endif
  endif

  light_s = .true.

  end subroutine set_light_strange_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_top_rcl

  ! This subroutine sets the top quark as light. This means that, if
  ! the top-quark is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the top-quark is massless this subroutine has no
  ! effect (i.e. top-quark collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(237).le.warning_limit) then
      warnings(237) = warnings(237) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 237: Processes already generated.'
      write(nx,*) '             The call of set_light_top_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(237)
    endif
  endif

  light_t = .true.

  end subroutine set_light_top_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_light_bottom_rcl

  ! This subroutine sets the bottom quark as light. This means that,
  ! if the bottom-quark is massive, its collinear singularities are
  ! regularized with its mass and terms proportional to the mass are
  ! discarded. If the bottom-quark is massless this subroutine has no
  ! effect (i.e. bottom-quark collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(238).le.warning_limit) then
      warnings(238) = warnings(238) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 238: Processes already generated.'
      write(nx,*) '             The call of set_light_bottom_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(238)
    endif
  endif

  light_b = .true.

  end subroutine set_light_bottom_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_electron_rcl

  ! This subroutine sets the electron as "not light". This means
  ! that, if the electron is massive, it is considered as heavy (i.e
  ! full mass dependence is kept). If the electron is massless this
  ! subroutine has no effect (i.e. electron collinear singularities
  ! are regularized dimensionally).

  if (processes_generated) then
    if (warnings(239).le.warning_limit) then
      warnings(239) = warnings(239) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 239: Processes already generated.'
      write(nx,*) '             The call of unset_light_electron_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(239)
    endif
  endif

  light_el = .false.

  end subroutine unset_light_electron_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_muon_rcl

  ! This subroutine sets the muon as "not light". This means that, if
  ! the muon is massive, it is considered as heavy (i.e full mass
  ! dependence is kept). If the muon is massless this subroutine has
  ! no effect (i.e. muon collinear singularities are regularized
  ! dimensionally).

  if (processes_generated) then
    if (warnings(240).le.warning_limit) then
      warnings(240) = warnings(240) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 240: Processes already generated.'
      write(nx,*) '             The call of unset_light_muon_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(240)
    endif
  endif

  light_mu = .false.

  end subroutine unset_light_muon_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_tau_rcl

  ! This subroutine sets the tau-lepton as "not light". This means
  ! that, if the tau-lepton is massive, it is considered as heavy (i.e
  ! full mass dependence is kept). If the tau-lepton is massless this
  ! subroutine has no effect (i.e. tau-lepton collinear singularities
  ! are regularized dimensionally).

  if (processes_generated) then
    if (warnings(241).le.warning_limit) then
      warnings(241) = warnings(241) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 241: Processes already generated.'
      write(nx,*) '             The call of unset_light_tau_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(241)
    endif
  endif

  light_ta = .false.

  end subroutine unset_light_tau_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_up_rcl

  ! This subroutine sets the up quark as "not light". This means
  ! that, if the up-quark is massive, it is considered as heavy (i.e
  ! full mass dependence is kept). If the up-quark is massless this
  ! subroutine has no effect (i.e. up-quark collinear singularities
  ! are regularized dimensionally).

  if (processes_generated) then
    if (warnings(242).le.warning_limit) then
      warnings(242) = warnings(242) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 242: Processes already generated.'
      write(nx,*) '             The call of unset_light_up_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(242)
    endif
  endif

  light_u = .false.

  end subroutine unset_light_up_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_down_rcl

  ! This subroutine sets the down quark as "not light". This means
  ! that, if the down-quark is massive, it is considered as heavy
  ! (i.e full mass dependence is kept). If the down-quark is massless
  ! this subroutine has no effect (i.e. down-quark collinear
  ! singularities are regularized dimensionally).

  if (processes_generated) then
    if (warnings(243).le.warning_limit) then
      warnings(243) = warnings(243) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 243: Processes already generated.'
      write(nx,*) '             The call of unset_light_down_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(243)
    endif
  endif

  light_d = .false.

  end subroutine unset_light_down_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_charm_rcl

  ! This subroutine sets the charm quark as "not light". This means
  ! that, if the charm-quark is massive, it is considered as heavy
  ! (i.e full mass dependence is kept). If the charm-quark is
  ! massless this subroutine has no effect (i.e. charm-quark
  ! collinear singularities are regularized dimensionally).

  if (processes_generated) then
    if (warnings(244).le.warning_limit) then
      warnings(244) = warnings(244) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 244: Processes already generated.'
      write(nx,*) '             The call of unset_light_charm_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(244)
    endif
  endif

  light_c = .false.

  end subroutine unset_light_charm_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_strange_rcl

  ! This subroutine sets the strange quark as "not light". This means
  ! that, if the strange-quark is massive, it is considered as heavy
  ! (i.e full mass dependence is kept). If the strange-quark is
  ! massless this subroutine has no effect (i.e. strange-quark
  ! collinear singularities are regularized dimensionally).

  if (processes_generated) then
    if (warnings(245).le.warning_limit) then
      warnings(245) = warnings(245) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 245: Processes already generated.'
      write(nx,*) '             The call of unset_light_strange_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(245)
    endif
  endif

  light_s = .false.

  end subroutine unset_light_strange_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_top_rcl

  ! This subroutine sets the top quark as "not light". This means
  ! that, if the top-quark is massive, it is considered as heavy
  ! (i.e full mass dependence is kept). If the top-quark is massless
  ! this subroutine has no effect (i.e. top-quark collinear
  ! singularities are regularized dimensionally).

  if (processes_generated) then
    if (warnings(246).le.warning_limit) then
      warnings(246) = warnings(246) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 246: Processes already generated.'
      write(nx,*) '             The call of unset_light_top_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(246)
    endif
  endif

  light_t = .false.

  end subroutine unset_light_top_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine unset_light_bottom_rcl

  ! This subroutine sets the bottom quark as "not light". This means
  ! that, if the bottom-quark is massive, it is considered as heavy
  ! (i.e full mass dependence is kept). If the bottom-quark is
  ! massless this subroutine has no effect (i.e. bottom-quark
  ! collinear singularities are regularized dimensionally).

  if (processes_generated) then
    if (warnings(247).le.warning_limit) then
      warnings(247) = warnings(247) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 247: Processes already generated.'
      write(nx,*) '             The call of unset_light_bottom_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(247)
    endif
  endif

  light_b = .false.

  end subroutine unset_light_bottom_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine use_dim_reg_soft_rcl

  ! This subroutine selects dimensional regularization for soft
  ! singularities.

  if (processes_generated) then
    if (warnings(248).le.warning_limit) then
      warnings(248) = warnings(248) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 248: Processes already generated.'
      write(nx,*) '             The call of use_dim_reg_soft_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(248)
    endif
  endif

  reg_soft = 1

  end subroutine use_dim_reg_soft_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine use_mass_reg_soft_rcl (m)

  ! This subroutine selects mass regularization for soft singularities
  ! The mass regulator is set to "m".
  ! REMARK: Mass regularization for soft singularities is allowed
  !         only in combination with mass regularization for
  !         collinear singularities of light charged fermions.
  !         Therefore no external charged massless fermions are
  !         allowed (and the code stops).

  real(dp), intent(in) :: m

  if (processes_generated) then
    if (warnings(249).le.warning_limit) then
      warnings(249) = warnings(249) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 249: Processes already generated.'
      write(nx,*) '             The call of use_mass_reg_soft_rcl ', &
                           'with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(249)
    endif
  endif

  if (m.lt.0d0) then
    if (warnings(250).le.warning_limit) then
      warnings(250) = warnings(250) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 250: use_mass_reg_soft_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(250)
    endif
    call istop (ifail,1)
  endif

  reg_soft = 2

  lambda  = m

  end subroutine use_mass_reg_soft_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_mass_reg_soft_rcl (m)

  ! This subroutine sets the mass regulator for soft singularities
  ! to "m".

  real(dp), intent(in) :: m

  if (processes_generated.and.dynamic_settings.eq.0) then
    if (warnings(251).le.warning_limit) then
      warnings(251) = warnings(251) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 251: Processes already generated.'
      write(nx,*) '             The call of use_mass_reg_soft_rcl ', &
                           'with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(251)
    endif
  endif

  if (m.lt.0d0) then
    if (warnings(252).le.warning_limit) then
      warnings(252) = warnings(252) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 252: use_mass_reg_soft_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(252)
    endif
    call istop (ifail,1)
  endif

  lambda  = m

  if (processes_generated) changed_lambda = .true.

  end subroutine set_mass_reg_soft_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_delta_uv_rcl (d)

  ! This subroutine sets coefficient "DeltaUV" of the UV pole
  ! (1/ep - EulerConstant + ln(4*pi)) in the MSbar scheme to "d".

  real(dp), intent(in) :: d

  if (processes_generated.and.dynamic_settings.eq.0) then
    if (warnings(253).le.warning_limit) then
      warnings(253) = warnings(253) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 253: Processes already generated.'
      write(nx,*) '            The call of set_delta_uv_rcl ', &
                           'with argument'
      write(nx,*) '             d = ',d
      write(nx,*) '            has no effects.'
      write(nx,*)
      call toomanywarnings(253)
    endif
  endif

  DeltaUV = d

  if (processes_generated) changed_DeltaUV = .true.

  end subroutine set_delta_uv_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_mu_uv_rcl (m)

  ! This subroutine sets the UV scale "muUV" to "m".

  real(dp), intent(in) :: m

  if (processes_generated.and.dynamic_settings.eq.0) then
    if (warnings(254).le.warning_limit) then
      warnings(254) = warnings(254) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 254: Processes already generated.'
      write(nx,*) '             The call of set_mu_uv_rcl with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(254)
    endif
  endif

  if (m.lt.0d0) then
    if (warnings(255).le.warning_limit) then
      warnings(255) = warnings(255) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 255: set_mu_uv_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(255)
    endif
    call istop (ifail,1)
  endif

  muUV = m

  if (processes_generated) changed_muUV = .true.

  end subroutine set_mu_uv_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_delta_ir_rcl (d,d2)

  ! This subroutine sets the coefficients "DeltaIR" of the IR
  ! single pole and "Deltair2" of the IR double pole to "d" and to
  ! "d2" respectively.

  real(dp), intent(in) :: d,d2

  if (processes_generated.and.dynamic_settings.eq.0) then
    if (warnings(256).le.warning_limit) then
      warnings(256) = warnings(256) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 256: Processes already generated.'
      write(nx,*) '             The call of set_delta_ir_rcl ', &
                           'with arguments'
      write(nx,*) '              d  = ',d
      write(nx,*) '              d2 = ',d2
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(256)
    endif
  endif

  DeltaIR  = d
  DeltaIR2 = d2

  if (processes_generated) changed_DeltaIR = .true.

  end subroutine set_delta_ir_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_mu_ir_rcl (m)

  ! This subroutine sets the IR scale "muIR" to "m".

  real(dp), intent(in) :: m

  if (processes_generated.and.dynamic_settings.eq.0) then
    if (warnings(257).le.warning_limit) then
      warnings(257) = warnings(257) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 257: Processes already generated.'
      write(nx,*) '             The call of set_mu_ir_rcl with argument'
      write(nx,*) '              m = ',m
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(257)
    endif
  endif

  if (m.lt.0d0) then
    if (warnings(258).le.warning_limit) then
      warnings(258) = warnings(258) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 258: set_mu_ir_rcl called with negative mass'
      write(nx,*)
      call toomanywarnings(258)
    endif
    call istop (ifail,1)
  endif

  muIR = m

  if (processes_generated) changed_muIR = .true.

  end subroutine set_mu_ir_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_alphas_rcl (a,s,Nf)

  ! This subroutine sets the values of alpha_s ("als") to "a", the
  ! renormalization scale "Qren" to "s" and the number of active
  ! quark flavours "Nfren" to "Nf".
  ! The choosen value of "a" should correspond to the physical value
  ! of alpha_s at the scale "s" computed in the "Nf"-flavour scheme.
  ! Accepted values for "Nf" are -1,3,4,5,6
  ! Nf = -1      -> all quarks with mass less than "s" are active
  !                 flavours
  ! Nf = 3,4,5,6 -> the 3,4,5,6 lightest quarks are the active
  !                 flavours

  integer,  intent(in) :: Nf
  real(dp), intent(in) :: a,s

  integer :: i

  if (a.le.0d0.or.s.le.0d0) then
    if (warnings(259).le.warning_limit) then
      warnings(259) = warnings(259) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 259: set_alphas_rcl called ', &
               'with wrong arguments'
      write(nx,*)
      call toomanywarnings(259)
    endif
    call istop (ifail,1)
  endif

  select case (Nf)
  case (-1,3,4,5,6)
  case default
    if (warnings(260).le.warning_limit) then
      warnings(260) = warnings(260) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'ERROR 260: set_alphas_rcl called ', &
                          'with a wrong number of '
      write(nx,*) '           light flavours Nf (accepted values are Nf = -1,3,4,5,6)'
      write(nx,*)
      call toomanywarnings(260)
    endif
    call istop (ifail,1)
  end select

  als   = a
  Qren  = s
  Nfren = Nf

  if (processes_generated) then
    ! set Nlq, the number of active flavours for alpha_s renormalization.
    select case (Nfren)
    case (-1)
      ! Nlq is the number of quarks lighter than Qren
      Nlq = 0
      do i = 1,6
        if (mq2(i).lt.Qren**2) Nlq = Nlq + 1
      enddo
    case (3,4,5)
      if (mq2(Nfren+1).eq.0d0) then
        if (warnings(261).le.warning_limit) then
          warnings(261) = warnings(261) + 1
          call openOutput
          write(nx,*)
          write(nx,*) 'ERROR 261: Wrong number Nfren of light flavours ', &
                             'for alphas renormaliztion'
          write(nx,*) '           (Nfren can not be smaller than the ', &
                             'number of massless quarks)'
          write(nx,*)
          call toomanywarnings(261)
        endif
        call istop (ifail,1)
      endif
      Nlq = Nfren
    case (6)
      Nlq = Nfren
    case default
      if (warnings(262).le.warning_limit) then
        warnings(262) = warnings(262) + 1
        call openOutput
        write(nx,*)
        write(nx,*) 'ERROR 262: Wrong value for the number Nfren of light ', &
                           'flavours for alphas renormaliztion'
        write(nx,*) '           (accepted values are Nfren = -1,3,4,5,6)'
        write(nx,*)
        call toomanywarnings(262)
      endif
      call istop (ifail,1)
    end select
  endif

  end subroutine set_alphas_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine get_alphas_rcl (a)

  ! This subroutine extracts the actual value of alpha_s ("als").

  real(dp), intent(out) :: a

  a = als

  end subroutine get_alphas_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine get_renormalization_scale_rcl (mu)

  ! This subroutine extracts the actual value of the renormalization
  ! scale for alpha_s ("Qren").

  real(dp), intent(out) :: mu

  mu = Qren

  end subroutine get_renormalization_scale_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine get_flavour_scheme_rcl (Nf)

  ! This subroutine extracts the actual value of the flavour-scheme
  ! for the renormalization of alpha_s ("Nf").

  integer, intent(out) :: Nf

  Nf = Nfren

  end subroutine get_flavour_scheme_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_alphas_masses_rcl (mc,mb,mt,gc,gb,gt)

  ! This subroutine allows to set the values of the squared quark
  ! masses for the running of alpha_s; their values are not used in
  ! the rest of the computation.
  ! The squared masses of the 3 heaviest quarks are set to
  ! "mc^2*sqrt(1+gc^2/mc^2)", "mb^2*sqrt(1+gb^2/mb^2)" and
  ! "mt^2*sqrt(1+gt^2/mt^2)" respectively.
  ! "mc", "mb", "mt" must satisfy 0 <= mc <= mb <= mt.
  ! The optional arguments "gc", "gb", "gt"  are ignored if the
  ! on-shell renormalization scheme is used.

  real(dp), intent(in)           :: mc,mb,mt
  real(dp), intent(in), optional :: gc,gb,gt

  if (processes_generated) then
    if (warnings(263).le.warning_limit) then
      warnings(263) = warnings(263) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 263: Processes already generated.'
      write(nx,*) '             The call of set_alphas_masses_rcl ', &
                           'with arguments'
      write(nx,*) '              mc = ',mc
      write(nx,*) '              mb = ',mb
      write(nx,*) '              mt = ',mt
      if (present(gc)) &
      write(nx,*) '              gc = ',gc
      if (present(gb)) &
      write(nx,*) '              gb = ',gb
      if (present(gt)) &
      write(nx,*) '              gt = ',gt
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(263)
    endif
  endif

  if ((mc.lt.0d0) .or. (mb .lt. mc) .or. (mt .lt. mb)) then
    if (warnings(264).le.warning_limit) then
      warnings(264) = warnings(264) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 264: set_alphas_masses_rcl called with wrong arguments.'
      write(nx,*) &
        '           Masses need to be positive and to satisfy 0 <= mc <= mb <= mt.'
      write(nx,*)
      call toomanywarnings(264)
    endif
    call istop (ifail,1)
  endif

  if (complex_mass_scheme .eq. 0 .and. &
      (present(gc) .or. present(gb) .or. present(gt))) then
    if (warnings(265).le.warning_limit) then
      warnings(265) = warnings(265) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'WARNING 265: Widths in set_alphas_masses_rcl are ignored'
      write(nx,*) &
        '             when complex_mass_scheme is 0.'
      write(nx,*)
      call toomanywarnings(265)
    endif
  end if

  mq(1:3) = 0d0

  if (present(gc)) then
    if (gc.lt.0d0) then
      if (warnings(266).le.warning_limit) then
        warnings(266) = warnings(266) + 1
        call openOutput
        write(nx,*)
        write(nx,*) &
          'ERROR 266: set_alphas_masses_rcl called with wrong arguments.'
        write(nx,*) &
          '           Widths can not be negative.'
        write(nx,*)
        call toomanywarnings(266)
      endif
      call istop (ifail,1)
    endif
    mq(4) = sqrt(abs(mc**2 - cId0*mc*gc))
  else
    mq(4) = mc
  endif

  if (present(gb)) then
    if (gb.lt.0d0) then
      if (warnings(267).le.warning_limit) then
        warnings(267) = warnings(267) + 1
        call openOutput
        write(nx,*)
        write(nx,*) &
          'ERROR 267: set_alphas_masses_rcl called with wrong arguments.'
        write(nx,*) &
          '           Widths can not be negative.'
        write(nx,*)
        call toomanywarnings(267)
      endif
      call istop (ifail,1)
    endif
    mq(5) = sqrt(abs(mb**2 - cId0*mb*gb))
  else
    mq(5)   = mb
  endif

  if (present(gt)) then
    if (gt.lt.0d0) then
      if (warnings(268).le.warning_limit) then
        warnings(268) = warnings(268) + 1
        call openOutput
        write(nx,*)
        write(nx,*) &
          'ERROR 268: set_alphas_masses_rcl called with wrong arguments.'
        write(nx,*) &
          '           Widths can not be negative.'
        write(nx,*)
        call toomanywarnings(268)
      endif
      call istop (ifail,1)
    endif
    mq(6) = sqrt(abs(mt**2 - cId0*mt*gt))
  else
    mq(6)   = mt
  endif

  use_active_qmasses = .true.

  end subroutine set_alphas_masses_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine use_gfermi_scheme_rcl (g,a,massesgf)

  ! This subroutine sets the EW renormalization scheme to the gfermi
  ! scheme (alpha is determined from gfermi) and, if present, sets
  ! the Fermi constant G_F ("gf") to "g" or alternatively sets,
  ! if present, alpha_GF ("algf") to "a".
  ! "massesgf", if present, selects the way W and Z masses are treated
  ! in the computation of alpha_GF (from "gf"):
  ! "massesgf = 0" -> alpha_GF = G_F*sqrt(2)/pi*Re(MW^2)*( 1 - Re(MW^2)/Re(MZ^2) )
  ! "massesgf = 1" -> alpha_GF = G_F*sqrt(2)/pi*abs( MW^2*( 1 - MW^2/MZ^2) )


  real(dp), optional, intent(in) :: g,a
  integer,  optional, intent(in) :: massesgf

  if (processes_generated) then
    if (warnings(269).le.warning_limit) then
      warnings(269) = warnings(269) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 269: Processes already generated.'
      write(nx,*) '             The call of use_gfermi_scheme_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(269)
    endif
  endif

  if (present(g).and.present(a)) then
    if (warnings(270).le.warning_limit) then
      warnings(270) = warnings(270) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 270: use_gfermi_scheme_rcl called with two arguments'
      write(nx,*)
      call toomanywarnings(270)
    endif
    call istop (ifail,1)
  endif

  if (present(g)) then
    if (g.lt.0d0) then
      if (warnings(271).le.warning_limit) then
        warnings(271) = warnings(271) + 1
        call openOutput
        write(nx,*)
        write(nx,*) &
          'ERROR 271: use_gfermi_scheme_rcl called with negative argument "g"'
        write(nx,*)
        call toomanywarnings(271)
      endif
      call istop (ifail,1)
    endif
  endif

  if (present(a)) then
    if (a.lt.0d0) then
      if (warnings(272).le.warning_limit) then
        warnings(272) = warnings(272) + 1
        call openOutput
        write(nx,*)
        write(nx,*) &
          'ERROR 272: use_gfermi_scheme_rcl called with negative argument "a"'
        write(nx,*)
        call toomanywarnings(272)
      endif
      call istop (ifail,1)
    endif
  endif

  ew_reno_scheme = 1

  if (present(g)) then
    gf   = g
    algf = - 1d0
  elseif (present(a)) then
    algf = a
  endif

  if (present(massesgf)) masses_in_gf_reno_scheme = massesgf

  end subroutine use_gfermi_scheme_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine use_alpha0_scheme_rcl (a)

  ! This subroutine sets the EW renormalization scheme to the alpha0
  ! scheme (alpha is renormalized at p^2 = 0) and, if present, sets
  ! alpha(0) ("al0") to "a".

  real(dp), optional, intent(in) :: a

  if (processes_generated) then
    if (warnings(273).le.warning_limit) then
      warnings(273) = warnings(273) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 273: Processes already generated.'
      write(nx,*) '             The call of use_alpha0_scheme_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(273)
    endif
  endif

  if (present(a)) then
    if (a.lt.0d0) then
      if (warnings(274).le.warning_limit) then
        warnings(274) = warnings(274) + 1
        call openOutput
        write(nx,*)
        write(nx,*) &
          'ERROR 274: use_alpha0_scheme_rcl called with negative argument'
        write(nx,*)
        call toomanywarnings(274)
      endif
      call istop (ifail,1)
    endif
  endif

  ew_reno_scheme = 2

  if (present(a)) al0 = a

  end subroutine use_alpha0_scheme_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine use_alphaz_scheme_rcl (a)

  ! This subroutine sets the EW renormalization scheme to the alphaz
  ! scheme (alpha is renormalized at p^2 = mz^2) and, if present, sets
  ! alpha(M_Z) ("alZ") to "a".

  real(dp), optional, intent(in) :: a

  if (processes_generated) then
    if (warnings(275).le.warning_limit) then
      warnings(275) = warnings(275) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 275: Processes already generated.'
      write(nx,*) '             The call of use_alphaz_scheme_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(275)
    endif
  endif

  if (present(a)) then
    if (a.lt.0d0) then
      if (warnings(276).le.warning_limit) then
        warnings(276) = warnings(276) + 1
        call openOutput
        write(nx,*)
        write(nx,*) &
          'ERROR 276: use_alphaz_scheme_rcl called with negative argument'
        write(nx,*)
        call toomanywarnings(276)
      endif
      call istop (ifail,1)
    endif
  endif

  ew_reno_scheme = 3

  if (present(a)) alZ = a

  end subroutine use_alphaz_scheme_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine get_alpha_rcl (a)

  ! This subroutine extract the actual value of alpha.

  real(dp), intent(out) :: a

  real(dp)    :: rew2,rez2
  complex(dp) :: mw2,mz2

  if (ew_reno_scheme.eq.1) then
    if (.not.processes_generated) then
      if (gf.lt.0d0) then
        alpha = algf
      elseif (masses_in_gf_reno_scheme.eq.0) then
        rew2 = mass_w**2
        rez2 = mass_z**2
        alpha = gf*sq2/pi*rew2*( 1d0 - rew2/rez2 )
      elseif (masses_in_gf_reno_scheme.eq.1) then
        mw2 = mass_w**2 - cId0*width_w*mass_w
        mz2 = mass_z**2 - cId0*width_z*mass_z
        alpha = gf*sq2/pi*abs( mw2*( 1d0 - mw2/mz2 ) )
      endif
    endif
    a = alpha
  elseif (ew_reno_scheme.eq.2) then
    a = al0
  elseif (ew_reno_scheme.eq.3) then
    a = alZ
  endif

  end subroutine get_alpha_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_complex_mass_scheme_rcl

  ! This subroutine selects the complex mass scheme of hep-ph/0505042
  ! for renormalization

  if (processes_generated) then
    if (warnings(277).le.warning_limit) then
      warnings(277) = warnings(277) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 277: Processes already generated.'
      write(nx,*) '             The call of ', &
                           'set_complex_mass_scheme_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(277)
    endif
  endif

  complex_mass_scheme = 1

  end subroutine set_complex_mass_scheme_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_on_shell_scheme_rcl

  ! This subroutine selects the on-shell scheme for renormalization.
  ! Just the real part of the self-energies are taken for the
  ! computation of the counterterms. The imaginary part of the masses
  ! are kept just in the denominator of propagators.

  if (processes_generated) then
    if (warnings(278).le.warning_limit) then
      warnings(278) = warnings(278) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 278: Processes already generated.'
      write(nx,*) '             The call of set_on_shell_scheme_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(278)
    endif
  endif

  complex_mass_scheme = 0

  end subroutine set_on_shell_scheme_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_resonant_particle_rcl (pa)

  ! This subroutine marks particle "pa" as resonant.
  ! "pa" can be the follwing characters:
  ! Scalars:       'H', 'p0', 'p+', 'p-'
  ! Vector bosons: 'g', 'A', 'Z', 'W+', 'W-'
  ! leptons:       'nu_e', 'nu_e~', 'e-', 'e+',
  !                'nu_mu', 'nu_mu~', 'mu-', 'mu+',
  !                'nu_tau', 'nu_tau~', 'tau-', 'tau+'
  ! quarks:        'u', 'u~', 'd', 'd~',
  !                'c', 'c~', 's', 's~',
  !                't', 't~', 'b', 'b~'

  character(len=*), intent(in) :: pa

  integer :: npa

  if (processes_generated) then
    if (warnings(279).le.warning_limit) then
      warnings(279) = warnings(279) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 279: Processes already generated.'
      write(nx,*) '             The call of set_resonant_particle_rcl ', &
                           'with argument'
      write(nx,*) '              pa = ',pa
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(279)
    endif
  endif

  npa = npar(pa)

  select case (npa)
  case (3,8,12,17)
    resPar(3)      = .true.
    resPar(8)      = .true.
    resPar(12)     = .true.
    resPar(17)     = .true.
  case (4,5,9,10,13,14,18,19)
    resPar(4:5)    = .true.
    resPar(9:10)   = .true.
    resPar(13:14)  = .true.
    resPar(18:19)  = .true.
  case (11)
    resPar(11)     = .true.
  case (23,35)
    resPar(23)     = .true.
    resPar(35)     = .true.
  case (24,36)
    resPar(24)     = .true.
    resPar(36)     = .true.
  case (25,37)
    resPar(25)     = .true.
    resPar(37)     = .true.
  case (26,38)
    resPar(26)     = .true.
    resPar(38)     = .true.
  case (27,39)
    resPar(27)     = .true.
    resPar(39)     = .true.
  case (28,40)
    resPar(28)     = .true.
    resPar(40)     = .true.
  case (29,41)
    resPar(29)     = .true.
    resPar(41)     = .true.
  case (30,42)
    resPar(30)     = .true.
    resPar(42)     = .true.
  case (31,43)
    resPar(31)     = .true.
    resPar(43)     = .true.
  case default ! never the case
    resPar(npa)    = .true.
  end select

  end subroutine set_resonant_particle_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_internal_projection_rcl (npr,v,p)

  ! Applies a projection p to the line v for process npr

  integer, intent(in) :: v,p,npr
  integer             :: pr,i,lmax,vp

  pr = 0
  do i = 1,prTot
    if (inpr(i).eq.npr) then
      pr = i; exit
    endif
  enddo

  lmax = legsIn(pr) + legsOut(pr)
  ! polproj stores only projections of the first 2**(lmax-1) momenta
  if (v .ge. 2**(lmax-1)) then
    vp = 2**(lmax) - 1 - v
    ! flip helicity for transversal polarization for opposite
    ! propagation direction
    if (abs(p) .eq. 1) then
      polproj(vp,pr) = -p
    else
      polproj(vp,pr) = p
    end if
  else
    vp = v
    polproj(vp,pr) = p
  end if

  end subroutine set_internal_projection_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_quarkline_rcl (npr,f1,f2)

  ! applies the constraint that the quark line between the quark f1 and f2
  ! remains connected through all currents.

  integer, intent(in) :: f1,f2,npr
  integer             :: pr,i

  pr = 0
  do i = 1,prTot
    if (inpr(i).eq.npr) then
      pr = i; exit
    endif
  enddo

  qflow(f1,pr) = f2
  qflow(f2,pr) = f1

  end subroutine set_quarkline_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine switchon_resonant_selfenergies_rcl

  ! This subroutine sets the value of "resSE" to .true.

  if (processes_generated) then
    if (warnings(280).le.warning_limit) then
      warnings(280) = warnings(280) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 280: Processes already generated.'
      write(nx,*) '             The call of ',                        &
                           'switchon_resonant_selfenergies_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(280)
    endif
  endif

  resSE = .true.

  end subroutine switchon_resonant_selfenergies_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine switchoff_resonant_selfenergies_rcl

  ! This subroutine sets the value of "resSE" to .false.

  if (processes_generated) then
    if (warnings(281).le.warning_limit) then
      warnings(281) = warnings(281) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 281: Processes already generated.'
      write(nx,*) '             The call of ',                         &
                           'switchoff_resonant_selfenergies_rcl ', &
                           'has no effects.'
      write(nx,*)
      call toomanywarnings(281)
    endif
  endif

  resSE = .false.

  end subroutine switchoff_resonant_selfenergies_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_dynamic_settings_rcl (n)

  ! This routine regulates the dynamic setting of parameters:
  ! n = 0 -> Just alpha_s can be set after generate_processes_rcl
  ! n = 1 -> Also lambda, DeltaUV, muUV, DeltaIR2, DeltaIR and muIR
  !          can be set after generate_processes_rcl

  integer, intent(in) :: n

  if (processes_generated) then
    if (warnings(282).le.warning_limit) then
      warnings(282) = warnings(282) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 282: Processes already generated.'
      write(nx,*) '             The call of set_dynamic_settings_rcl ', &
                           'with arguments'
      write(nx,*) '              n = ',n
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(282)
    endif
  endif

  dynamic_settings = n

  end subroutine set_dynamic_settings_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_momenta_correction_rcl (mc)

  ! This routine regulates the correction of external momenta given
  ! by the user, when on-shell conditions or momentum conservation are
  ! not fulfilled.
  ! mc = .true.  -> recola tries to correct the momenta
  ! mc = .false. -> recola does not try to correct the momenta
  ! When the correction does not work or is turned off, all
  ! computed amplitudes with bad momenta are set to zero.

  logical, intent(in) :: mc

  momenta_correction = mc

  end subroutine set_momenta_correction_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_draw_level_branches_rcl (n)

  ! This subroutine sets the options for the generation of a .tex
  ! file, to draw the off-shell currents of the processes:
  ! n = 0 -> the .tex file is not generated.
  ! n = 1 -> the .tex file is generated without colour structures.
  ! n = 2 -> the .tex file is generated with colour structures.

  integer, intent(in) :: n

  if (processes_generated) then
    if (warnings(283).le.warning_limit) then
      warnings(283) = warnings(283) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 283: Processes already generated.'
      write(nx,*) '             The call of ', &
                           'set_draw_level_branches_rcl ', &
                           'with argument'
      write(nx,*) '              n = ',n
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(283)
    endif
  endif

  if (n.lt.0.or.n.gt.2) then
    if (warnings(284).le.warning_limit) then
      warnings(284) = warnings(284) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 284: set_draw_level_branches_rcl called with wrong argument'
      write(nx,*)
      call toomanywarnings(284)
    endif
    call istop (ifail,1)
  endif

  draw = n

  end subroutine set_draw_level_branches_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_print_level_amplitude_rcl (n)

  ! This subroutine sets the level for printing the amplitude to "n":
  ! n = 0 -> nothing is printed.
  ! n = 1 -> The non vanishining amplitudes A(c,h) are printed at LO
  !          and NLO (if NLO is computed) for each coulour structure
  !          "c" and helicity "h"
  ! n = 2 -> In addition to n=1, the NLO (if computed) contributions
  !          to A(c,h) from bare-loop, counterterms and rational part
  !          are also printed.

  integer, intent(in) :: n

  if (n.lt.0.or.n.gt.2) then
    if (warnings(285).le.warning_limit) then
      warnings(285) = warnings(285) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 285: set_print_level_amplitude_rcl called with wrong argument'
      write(nx,*)
      call toomanywarnings(285)
    endif
    call istop (ifail,1)
  endif

  writeMat = n

  end subroutine set_print_level_amplitude_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_print_level_squared_amplitude_rcl (n)

  ! This subroutine sets the level for printing the squared amplitude
  ! to "n":
  ! n = 0 -> nothing is printed.
  ! n = 1 -> The non vanishining squared amplitudes A2
  !          summed/averaged over colour and helicities are
  !          printed at LO and NLO (if NLO is computed).
  ! n = 2 -> In addition to n=1, the NLO (if computed) contributions
  !          to A2 from bare-loop, counterterms and rational part
  !          are also printed.
  ! n = 3 -> In addition to n=2, also the squared amplitudes A2h
  !          summed/averaged over colour are printed at LO and NLO
  !          (if NLO is computed) for each helicity "h".

  integer, intent(in) :: n

  if (n.lt.0.or.n.gt.3) then
    if (warnings(286).le.warning_limit) then
      warnings(286) = warnings(286) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 286: set_print_level_squared_amplitude_rcl called with wrong argument'
      write(nx,*)
      call toomanywarnings(286)
    endif
    call istop (ifail,1)
  endif

  writeMat2 = n

  end subroutine set_print_level_squared_amplitude_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_print_level_correlations_rcl (n)

  ! This subroutine sets the level for printing colour correlation
  ! (if computed) to "n":
  ! n = 0 -> nothing is printed.
  ! n = 1 -> The non vanishining colour-correlated Born squared
  !          amplitudes A2c(n,m) are printed for each emitter "n" and
  !          spectator "m".

  integer, intent(in) :: n

  if (n.lt.0.or.n.gt.1) then
    if (warnings(287).le.warning_limit) then
      warnings(287) = warnings(287) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 287: set_print_level_correlations_rcl called with wrong argument'
      write(nx,*)
      call toomanywarnings(287)
    endif
    call istop (ifail,1)
  endif

  writeCor = n

  end subroutine set_print_level_correlations_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_print_level_RAM_rcl (n)

  ! This subroutine sets to "n" the level for printing RAM needed in
  ! the present run of Recola:
  ! n = 0 -> nothing is printed.
  ! n = 1 -> RAM permanently used by process generation and process
  !          computation is printed.
  ! n = 2 -> In addition to n=1, RAM temporally used by process
  !          generation is printed.

  integer, intent(in) :: n

  if (processes_generated) then
    if (warnings(288).le.warning_limit) then
      warnings(288) = warnings(288) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 288: Processes already generated.'
      write(nx,*) '             The call of set_print_level_RAM_rcl ', &
                           'with argument'
      write(nx,*) '              n = ',n
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(288)
    endif
  endif

  if (n.lt.0.or.n.gt.2) then
    if (warnings(289).le.warning_limit) then
      warnings(289) = warnings(289) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 289: set_print_level_RAM_rcl called with wrong argument'
      write(nx,*)
      call toomanywarnings(289)
    endif
    call istop (ifail,1)
  endif

  writeRAM = n

  end subroutine set_print_level_RAM_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine scale_coupling3_rcl (fac,pa1,pa2,pa3)

  ! This subroutine scales the 3-particle coupling between particles
  ! "pa1,pa2,pa3" by a factor "fac".
  ! "pa1,pa2,pa3" can be the follwing characters:
  ! Scalars:       'H', 'p0', 'p+', 'p-'
  ! Vector bosons: 'g', 'A', 'Z', 'W+', 'W-'
  ! leptons:       'nu_e', 'nu_e~', 'e-', 'e+',
  !                'nu_mu', 'nu_mu~', 'mu-', 'mu+',
  !                'nu_tau', 'nu_tau~', 'tau-', 'tau+'
  ! quarks:        'u', 'u~', 'd', 'd~',
  !                'c', 'c~', 's', 's~',
  !                't', 't~', 'b', 'b~'
  ! REMARK: the rescaling on the counterterms and rational terms is
  !         not yet implemented

  complex(dp), intent(in) :: fac
  character(len=*), intent(in) :: pa1,pa2,pa3

  integer :: npa1,npa2,npa3

  if (processes_generated) then
    if (warnings(290).le.warning_limit) then
      warnings(290) = warnings(290) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 290: Processes already generated.'
      write(nx,*) '             The call of scale_coupling3_rcl ', &
                           'with arguments'
      write(nx,*) '              fac = ',fac
      write(nx,*) '              pa1 = ',pa1
      write(nx,*) '              pa2 = ',pa2
      write(nx,*) '              pa3 = ',pa3
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(290)
    endif
  endif

  npa1 = npar(pa1)
  npa2 = npar(pa2)
  npa3 = npar(pa3)

  coupling3(npa1,npa2,npa3) = fac
  coupling3(npa1,npa3,npa2) = fac
  coupling3(npa2,npa1,npa3) = fac
  coupling3(npa2,npa3,npa1) = fac
  coupling3(npa3,npa1,npa2) = fac
  coupling3(npa3,npa2,npa1) = fac

  end subroutine scale_coupling3_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine scale_coupling4_rcl (fac,pa1,pa2,pa3,pa4)

  ! This subroutine scales the 4-particle coupling between particles
  ! "pa1,pa2,pa3,pa4" by a factor "fac".
  ! "pa1,pa2,pa3,pa4" can be the follwing characters:
  ! Scalars:       'H', 'p0', 'p+', 'p-'
  ! Vector bosons: 'g', 'A', 'Z', 'W+', 'W-'
  ! leptons:       'nu_e', 'nu_e~', 'e-', 'e+',
  !                'nu_mu', 'nu_mu~', 'mu-', 'mu+',
  !                'nu_tau', 'nu_tau~', 'tau-', 'tau+'
  ! quarks:        'u', 'u~', 'd', 'd~',
  !                'c', 'c~', 's', 's~',
  !                't', 't~', 'b', 'b~'
  ! REMARK: the rescaling on the counterterms and rational terms is
  !         not yet implemented

  complex(dp), intent(in) :: fac
  character(len=*), intent(in) :: pa1,pa2,pa3,pa4

  integer :: npa1,npa2,npa3,npa4

  if (processes_generated) then
    if (warnings(291).le.warning_limit) then
      warnings(291) = warnings(291) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 291: Processes already generated.'
      write(nx,*) '             The call of scale_coupling4_rcl ', &
                           'with arguments'
      write(nx,*) '              fac = ',fac
      write(nx,*) '              pa1 = ',pa1
      write(nx,*) '              pa2 = ',pa2
      write(nx,*) '              pa3 = ',pa3
      write(nx,*) '              pa4 = ',pa4
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(291)
    endif
  endif

  npa1 = npar(pa1)
  npa2 = npar(pa2)
  npa3 = npar(pa3)
  npa4 = npar(pa4)

  coupling4(npa1,npa2,npa3,npa4) = fac
  coupling4(npa1,npa2,npa4,npa3) = fac
  coupling4(npa1,npa3,npa2,npa4) = fac
  coupling4(npa1,npa3,npa4,npa2) = fac
  coupling4(npa1,npa4,npa2,npa3) = fac
  coupling4(npa1,npa4,npa3,npa2) = fac
  coupling4(npa2,npa1,npa3,npa4) = fac
  coupling4(npa2,npa1,npa4,npa3) = fac
  coupling4(npa2,npa3,npa1,npa4) = fac
  coupling4(npa2,npa3,npa4,npa1) = fac
  coupling4(npa2,npa4,npa1,npa3) = fac
  coupling4(npa2,npa4,npa3,npa1) = fac
  coupling4(npa3,npa1,npa2,npa4) = fac
  coupling4(npa3,npa1,npa4,npa2) = fac
  coupling4(npa3,npa2,npa1,npa4) = fac
  coupling4(npa3,npa2,npa4,npa1) = fac
  coupling4(npa3,npa4,npa1,npa2) = fac
  coupling4(npa3,npa4,npa2,npa1) = fac
  coupling4(npa4,npa1,npa2,npa3) = fac
  coupling4(npa4,npa1,npa3,npa2) = fac
  coupling4(npa4,npa2,npa1,npa3) = fac
  coupling4(npa4,npa2,npa3,npa1) = fac
  coupling4(npa4,npa3,npa1,npa2) = fac
  coupling4(npa4,npa3,npa2,npa1) = fac

  end subroutine scale_coupling4_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine switchoff_coupling3_rcl (pa1,pa2,pa3)

  ! This subroutine switches off the 3-particle coupling between
  ! particles "pa1,pa2,pa3" .
  ! "pa1,pa2,pa3" can be the follwing characters:
  ! Scalars:       'H', 'p0', 'p+', 'p-'
  ! Vector bosons: 'g', 'A', 'Z', 'W+', 'W-'
  ! leptons:       'nu_e', 'nu_e~', 'e-', 'e+',
  !                'nu_mu', 'nu_mu~', 'mu-', 'mu+',
  !                'nu_tau', 'nu_tau~', 'tau-', 'tau+'
  ! quarks:        'u', 'u~', 'd', 'd~',
  !                'c', 'c~', 's', 's~',
  !                't', 't~', 'b', 'b~'
  ! REMARK: the switching off on the counterterms and rational terms is
  !         not yet implemented

  character(len=*), intent(in) :: pa1,pa2,pa3

  integer :: npa1,npa2,npa3

  if (processes_generated) then
    if (warnings(292).le.warning_limit) then
      warnings(292) = warnings(292) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 292: Processes already generated.'
      write(nx,*) '             The call of switchoff_coupling3_rcl ', &
                           'with arguments'
      write(nx,*) '              pa1 = ',pa1
      write(nx,*) '              pa2 = ',pa2
      write(nx,*) '              pa3 = ',pa3
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(292)
    endif
  endif

  npa1 = npar(pa1)
  npa2 = npar(pa2)
  npa3 = npar(pa3)

  coupling3(npa1,npa2,npa3) = 0d0
  coupling3(npa1,npa3,npa2) = 0d0
  coupling3(npa2,npa1,npa3) = 0d0
  coupling3(npa2,npa3,npa1) = 0d0
  coupling3(npa3,npa1,npa2) = 0d0
  coupling3(npa3,npa2,npa1) = 0d0

  end subroutine switchoff_coupling3_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine switchoff_coupling4_rcl (pa1,pa2,pa3,pa4)

  ! This subroutine switches off the 4-particle coupling between
  ! particles "pa1,pa2,pa3,pa4".
  ! "pa1,pa2,pa3,pa4" can be the follwing characters:
  ! Scalars:       'H', 'p0', 'p+', 'p-'
  ! Vector bosons: 'g', 'A', 'Z', 'W+', 'W-'
  ! leptons:       'nu_e', 'nu_e~', 'e-', 'e+',
  !                'nu_mu', 'nu_mu~', 'mu-', 'mu+',
  !                'nu_tau', 'nu_tau~', 'tau-', 'tau+'
  ! quarks:        'u', 'u~', 'd', 'd~',
  !                'c', 'c~', 's', 's~',
  !                't', 't~', 'b', 'b~'
  ! REMARK: the switching off on the counterterms and rational terms
  !         is not yet implemented

  character(len=*), intent(in) :: pa1,pa2,pa3,pa4

  integer :: npa1,npa2,npa3,npa4

  if (processes_generated) then
    if (warnings(293).le.warning_limit) then
      warnings(293) = warnings(293) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 293: Processes already generated.'
      write(nx,*) '             The call of switchoff_coupling4_rcl ', &
                           'with arguments'
      write(nx,*) '              pa1 = ',pa1
      write(nx,*) '              pa2 = ',pa2
      write(nx,*) '              pa3 = ',pa3
      write(nx,*) '              pa4 = ',pa4
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(293)
    endif
  endif

  npa1 = npar(pa1)
  npa2 = npar(pa2)
  npa3 = npar(pa3)
  npa4 = npar(pa4)

  coupling4(npa1,npa2,npa3,npa4) = 0d0
  coupling4(npa1,npa2,npa4,npa3) = 0d0
  coupling4(npa1,npa3,npa2,npa4) = 0d0
  coupling4(npa1,npa3,npa4,npa2) = 0d0
  coupling4(npa1,npa4,npa2,npa3) = 0d0
  coupling4(npa1,npa4,npa3,npa2) = 0d0
  coupling4(npa2,npa1,npa3,npa4) = 0d0
  coupling4(npa2,npa1,npa4,npa3) = 0d0
  coupling4(npa2,npa3,npa1,npa4) = 0d0
  coupling4(npa2,npa3,npa4,npa1) = 0d0
  coupling4(npa2,npa4,npa1,npa3) = 0d0
  coupling4(npa2,npa4,npa3,npa1) = 0d0
  coupling4(npa3,npa1,npa2,npa4) = 0d0
  coupling4(npa3,npa1,npa4,npa2) = 0d0
  coupling4(npa3,npa2,npa1,npa4) = 0d0
  coupling4(npa3,npa2,npa4,npa1) = 0d0
  coupling4(npa3,npa4,npa1,npa2) = 0d0
  coupling4(npa3,npa4,npa2,npa1) = 0d0
  coupling4(npa4,npa1,npa2,npa3) = 0d0
  coupling4(npa4,npa1,npa3,npa2) = 0d0
  coupling4(npa4,npa2,npa1,npa3) = 0d0
  coupling4(npa4,npa2,npa3,npa1) = 0d0
  coupling4(npa4,npa3,npa1,npa2) = 0d0
  coupling4(npa4,npa3,npa2,npa1) = 0d0

  end subroutine switchoff_coupling4_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_ifail_rcl (i)

  ! This subroutine sets the ifail flag to "i".
  ! The ifail flag regulates the behaviour of Recola in case of error.
  ! On input:
  ! ifail = -1 -> the code does not stop, when an error occurs
  ! ifail =  0 -> the code stops, when an error occurs

  integer, intent(in) :: i

  if (i.ne.0.and.i.ne.-1) then
    call openOutput
    write(nx,*)
    write(nx,*) &
      'ERROR: set_ifail_rcl called with wrong argument'
    write(nx,*)
    stop
  endif

  ifail = i

  end subroutine set_ifail_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine get_ifail_rcl (i)

  ! This subroutine extracts the actual value of the ifail flag in the
  ! output argument "i".
  ! The ifail flag regulates the behaviour of Recola in case of error.
  ! On output:
  ! ifail = 0 -> no errors occured
  ! ifail = 1 -> an error occured because of a wrong call of
  !              subroutines by the user
  ! ifail = 2 -> an error occured because of an internal bug of Recola

  integer, intent(out) :: i

  if (ifail.le.0) then; i = 0
  else;                 i = ifail
  endif

  end subroutine get_ifail_rcl


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_collier_mode_rcl (mode)
    use collier, only: SetMode_cll

    ! This subroutine sets the collier mode to "mode".
    ! 1 -> COLI
    ! 2 -> DD
    ! 3 -> COLI+DD

    integer, intent(in) :: mode

    if (mode .lt. 1 .or. mode .gt. 3) then
      if (warnings(294).le.warning_limit) then
        warnings(294) = warnings(294) + 1
        call openOutput
        write(nx,*)
        write(nx,*) 'WARNING 294: The given collier mode is not supported.'
        write(nx,*) '             mode = ',mode
        write(nx,*) '             Input will be ignored.'
        write(nx,*)
        call toomanywarnings(294)
      endif
    else
      collier_mode = mode

      if (processes_generated) then
        call SetMode_cll(collier_mode)
      end if
    endif

  end subroutine set_collier_mode_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_collier_output_dir_rcl (dir)

  ! This subroutine sets the collier output directory to "dir".
  ! With "dir" = 'default', the output directory of collier will be the
  ! one choosen by collier per default.
  ! For all other values, the output directory of collier will be set
  ! to 'dir', where 'dir' can be a relative or absolute path.

  character(*), intent(in) :: dir

  if (processes_generated) then
    if (warnings(295).le.warning_limit) then
      warnings(295) = warnings(295) + 1
      call openOutput
      write(nx,*)
      write(nx,*) 'WARNING 295: Processes already generated.'
      write(nx,*) '             The call of set_collier_output_dir_rcl ', &
                           'with argument'
      write(nx,*) '              dir = ',dir
      write(nx,*) '             has no effects.'
      write(nx,*)
      call toomanywarnings(295)
    endif
  endif

  if (len(trim(dir)).gt.999) then
    if (warnings(296).le.warning_limit) then
      warnings(296) = warnings(296) + 1
      call openOutput
      write(nx,*)
      write(nx,*) ' ERROR 296:'
      write(nx,*) ' the name of collier output directory'
      write(nx,*)
      write(nx,'(a)') trim(dir)
      write(nx,*)
      write(nx,*) ' is too long.'
      write(nx,*)
      call toomanywarnings(296)
    endif
    call istop (ifail,1)
  endif

  collier_output_dir = trim(dir)

  end subroutine set_collier_output_dir_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_compute_ir_poles_rcl (mode)

  integer, intent(in) :: mode

  if (mode .lt. 0 .or. mode .gt. 1) then
    if (warnings(297).le.warning_limit) then
      warnings(297) = warnings(297) + 1
      call openOutput
      write(nx,*)
      write(nx,*) &
        'ERROR 297: set_compute_ir_poles_rcl called with wrong argument'
      write(nx,*)
      call toomanywarnings(297)
    endif
    call istop (ifail,1)
  endif

  compute_ir_poles = mode

  end subroutine set_compute_ir_poles_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_output_file_rcl (x)

  ! This subroutine sets the output file name to "x".
  ! Any output produced by recola is written to the file named "x",
  ! which is created with the first call of set_output_file_rcl with
  ! argument "x" (if it already exists, its content is cleared).
  ! Any output produced by recola is written to the file named "x",
  ! which is created with the first call of set_output_file_rcl with
  ! argument "x" (if it already exists, its content is cleared).
  ! By a subsequent call of set_output_file_rcl with a different
  ! argument "y", the file "x" remains open (until the call
  ! reset_recola_rcl), but the output is now printed on "y".
  ! If the set_output_file_rcl is again called with argument
  ! "x", further output is then appended to the open file "x".
  ! By the call of reset_recola_rcl all opened output files are
  ! closed and, by further call of set_output_file_rcl with argument
  ! "x", the content of the existing file "x" is cleared.

  character(*), intent(in) :: x

  if (len(trim(x)).gt.999) then
    if (warnings(298).le.warning_limit) then
      warnings(298) = warnings(298) + 1
      write(*,*)
      write(*,*) ' WARNING 298:'
      write(*,*) ' the name of output-file'
      write(*,*)
      write(*,'(a)') trim(x)
      write(*,*)
      write(*,*) ' is too long and has been truncated to'
      write(*,*)
      write(*,'(a999)') trim(x)
      write(*,*)
      call toomanywarnings(298)
    endif
  endif

  outputfile = trim(x)

  if (recolascreen.eq.1) then
    write(*,*)
    write(*,'(1x,75("x"))')
    write(*,*)
    if (trim(outputfile).eq.'*') then
      write(*,*) 'From now on the output of RECOLA is written ', &
                 'on the screen'
    else
      write(*,*) 'From now on the output of RECOLA is written ', &
                 'to the file ',trim(outputfile)
    endif
    write(*,*)
    write(*,'(1x,75("x"))')
    write(*,*)
  endif

  end subroutine set_output_file_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  end module input_rcl

!#####################################################################
