!#####################################################################
!!
!!  File  outgoing_momenta_rcl.f90
!!  is part of RECOLA (REcursive Computation of One Loop Amplitudes)
!!
!!  Copyright (C) 2015-2019   Stefano Actis, Ansgar Denner,
!!                            Lars Hofer, Jean-Nicolas Lang,
!!                            Andreas Scharf, Sandro Uccirati
!!
!!  RECOLA is licenced under the GNU GPL version 3,
!!         see COPYING for details.
!!
!#####################################################################

  module outgoing_momenta_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  use input_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  implicit none

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  contains

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  subroutine set_outgoing_momenta_rcl (npr,pIn,p)

  ! This subroutine sets the momenta for the process with process
  ! number "npr". "pIn" are the incoming momenta.
  ! "p" are all momenta (incoming and outgoing) as generated by the
  ! subroutine (of course the incoming momenta coincide with "pIn").
  ! The subroutine uses RAMBO to generate the momenta.

  integer,  intent(in)  :: npr
  real(dp), intent(in)  :: pIn(0:,:)
  real(dp), intent(out) :: p(0:,:)

#ifdef WITHRAMBO
  integer               :: pr,i,lin,lout,l
  real(dp), allocatable :: pOutR(:,:),m(:)
  real(dp)              :: roots,wt,pCM(0:3),beta(1:3),beta2,gamma, &
                           E,pv(1:3),betap

  pr = 0
  do i = 1,prTot
    if (inpr(i).eq.npr) then
      pr = i; exit
    endif
  enddo
  if (pr.eq.0) then
    if (warnings(386).le.warning_limit) then
      warnings(386) = warnings(386) + 1
      call openOutput
      write(nx,*)
      write(nx,'(a,i3)') ' ERROR 386: set_outgoing_momenta_rcl called '// &
                                 'with undefined process index ',npr
      write(nx,*)
      call toomanywarnings(386)
    endif
    call istop (ifail,1)
  endif

  lin  =  legsIn(pr)
  lout = legsOut(pr)
  l    = lin + lout

  p(0:3,1:lin) = pIn(0:3,1:lin)

  if (lout.eq.1) then

    p(0:3,lin+1) = pIn(0:3,1) + pIn(0:3,2)

  else

    do i = 0,3
      pCM(i) = sum(pIn(i,1:lin))
    enddo
    roots = sqrt(pCM(0)**2-sum(pCM(1:3)**2))
    beta(1:3) = pCM(1:3)/pCM(0)
    beta2 = sum(beta(1:3)**2)
    gamma = pCM(0)/roots

    allocate (m(l))

    do i = 1,l
      m(i) = mONS(newleg(i,pr),pr)
    enddo

    allocate (pOutR(4,lout))

    call rambo(lout,roots,m(lin+1:l),pOutR,wt,0)

    if (beta2.eq.0d0) then
      p(0,lin+1:l) = pOutR(4,1:lout)
      p(1,lin+1:l) = pOutR(1,1:lout)
      p(2,lin+1:l) = pOutR(2,1:lout)
      p(3,lin+1:l) = pOutR(3,1:lout)
    else
      do i = 1,lout
        E = pOutR(4,i)
        pv(1:3) = pOutR(1:3,i)
        betap = sum(beta(:)*pv(:))
        p(0,lin+i) = gamma*( E + betap )
        p(1:3,lin+i) = &
        pv(1:3) + beta(1:3)*( gamma*E - betap/beta2*(1d0-gamma) )
      enddo
    endif

    deallocate (pOutR)

!    call openOutput
!    do i = 1,l
!      write(nx,'(3x,a,i1,a,1x,"(",3(f16.10,","),f16.10,")")') &
!        'p',i,' =',p(0,i),p(1,i),p(2,i),p(3,i)
!    enddo
!    write(nx,'(1x,a,1x,"(",3(f16.10,","),f16.10,")")') &
!      'pIN  =', &
!      sum(p(0,1:lin)), &
!      sum(p(1,1:lin)), &
!      sum(p(2,1:lin)), &
!      sum(p(3,1:lin))
!    write(nx,'(1x,a,1x,"(",3(f16.10,","),f16.10,")")') &
!      'pOUT =', &
!      sum(p(0,lin+1:l)), &
!      sum(p(1,lin+1:l)), &
!      sum(p(2,lin+1:l)), &
!      sum(p(3,lin+1:l))
!    do i = 1,l
!      write(nx,'(1x,a,i1,a,1x,f16.10,10x,a,i1,a,1x,f16.10)') &
!        'p',i,'^2 =',p(0,i)**2 - p(1,i)**2 - p(2,i)**2 - p(3,i)**2, &
!        'm',i,'^2 =',m(i)**2
!    enddo
!    write(nx,*)

    deallocate (m)
  end if
#else
  if (warnings(387).le.warning_limit) then
    warnings(387) = warnings(387) + 1
    call openOutput
    write(nx,*)
    write(nx,'(a,i3)') ' ERROR 387: Recola not compiled with RAMBO. '// &
                               'Cannot call set_outgoing_momenta_rcl.'
    write(nx,*)
    call toomanywarnings(387)
  endif
#endif

  end subroutine set_outgoing_momenta_rcl

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  end module outgoing_momenta_rcl

!#####################################################################



